%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%    Modified from http_util
%   by MultimediaN project
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(http_util,
	  [ reply_file/1,		% +File
	    reply_file/2,		% +File, +Request
	    reply_file_vars/2		% +File, +Vars
	    ]).

:- use_module(library('http/http_open')).
:- use_module(library('http/html_write')).	  
:- use_module(library('http/mimetype')).
:- use_module(library('http/http_header')).

   

reply_file(File) :-
	reply_file(File, []).

reply_file(File, Request) :-
	absolute_file_name(File, Path,
			   [ extensions(['', in]),
			     access(read)
			   ]), !,
	(   memberchk(if_modified_since(Since), Request),
	    time_file(Path, Time),
	    catch(http_timestamp(Time, Since), _, fail)
	->  throw(http_reply(not_modified))
	;   file_name_extension(Base, in, Path)
	->  file_mime_type(Base, MimeType),
	    setting(prefix(Prefix)),
	    reply_file_vars(Path,
			    [ prefix=Prefix
			    ],
			    [ static(true),
			      mime_type(MimeType)
			    ])
	;   file_mime_type(Path, MimeType),
	    throw(http_reply(file(MimeType, Path)))
	).


%	reply_file_vars(+File, +Var=Value..., +Options)
%	
%	Reply a file.  Files may refer to variables using the contruct
%	@varname@.  Defined options are:
%	
%		* static(true)
%		This implies the variables do not change during the
%		lifetime of the server.
%		
%		* request(Request)
%		Pass the request info.  In present and the above is
%		present this routine will do last-modified processing.
%		
%		* mime_type(Type)
%		Bypass default mimetype calculation

reply_file_vars(File, Bindings) :-
	reply_file_vars(File, Bindings, []).

reply_file_vars(File, Bindings, Options) :-
	absolute_file_name(File, Path, [access(read)]),
	modified_header(Path, Options),
	(   memberchk(mime_type(MimeType), Options)
	->  true
	;   file_mime_type(Path, MimeType)
	),
	open(Path, read, In),
	format('Content-type: ~w~n~n', [MimeType]),
	get_code(In, C0),
	copy_with_vars(C0, In, Bindings),
	close(In).

modified_header(Path, Options) :-
	memberchk(static(true), Options),
	memberchk(request(Request), Options), !,
	time_file(Path, Time),
	http_timestamp(Time, RfcTime),
	(   memberchk(request(Request), Options),
	    memberchk(if_modified_since(RfcTime), Request)
	->  throw(http_reply(not_modified))
	;   format('Last-modified: ~w~n', [RfcTime])
	).
modified_header(Path, Options) :-
	memberchk(static(true), Options), !,
	time_file(Path, Time),
	http_timestamp(Time, RfcTime),
	format('Last-modified: ~w~n', [RfcTime]).
modified_header(_, _).


copy_with_vars(-1, _, _) :- !.
copy_with_vars(0'@, In, Bindings) :- !,
	get_code(In, C0),
	read_var_name(C0, In, VarNameS, C1),
	atom_codes(VarName, VarNameS),
	(   C1 == 0'@,
	    memberchk(VarName=Value, Bindings)
	->  write(Value),
	    get_code(In, C2)
	;   format('@~w', [VarName]),
	    C2 = C1
	),
	copy_with_vars(C2, In, Bindings).

copy_with_vars(C0, In, Bindings) :-
	put_code(C0),
	get_code(In, C1),
	copy_with_vars(C1, In, Bindings).

read_var_name(C0, In, [C0|T], End) :-
	code_type(C0, alpha), !,
	get_code(In, C1),
	read_var_name(C1, In, T, End).
read_var_name(C0, _In, [], C0).
	

