/*  $Id: pion_selection.pl,v 1.12 2006/10/17 10:11:10 huang Exp $


    Author:        Zhisheng Huang
    E-mail:        huang@cs.vu.nl
    WWW:           http://www.cs.vu.nl/~huang
    Copyright (C): 2004-2005 Vrije University Amsterdam

This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    As a special exception, if you link this library with other files,
    compiled with a Free Software compiler, to produce an executable, this
    library does not by itself cause the resulting executable to be covered
    by the GNU General Public License. This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.

*/

:- module(pion_selection,
	  [pion_selection_function/5, %+SelectionFunctionType,  +QueryElement, +WorkingSet, +KBID, -SelectedSet
	   pion_selection_function/10,%+Type, +KBID, -Query, +K, +CurrentWorkingSet, +ICRk1, -NewSelectedSet, -NewWorkingSet, -ICRk, -Status
	   pion_icr_set_element/2,         %+Element, -ICR
         pion_icr_set/2,         		%+Elements, -ICR
         pion_direct_concept_relevant/2,	%+Element, +Element or ICR	   
         selected_set/5,
	   google_distance_element/4,      %+Type:google_distance_max/min/sum, +NameList, +CoreKeyword, -Distance
	   pion_relevant/2			%+Element, +Element or ICR
	  ]).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.2.4 added
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% sf(KBID, Query, K)=\Sigma' with respect to +Type
% \Sigma = CurrentWorkingSet = sf(KBID, Query, (K-1)), \Sigma'=\Sigma + NewSelectedSet=NewWorkingSet 
% Status =yes: \Sigma' is consistent, no: inconsistent, 
%pion_selection_function(+Type, +KBID, +Query, +K, +CurrentWorkingSet, +ICRk1, -NewSelectedSet, -NewWorkingSet, -ICRk, -Status)

pion_selection_function(Type, KBID, Q, K, _, _, L, L, ICRk, Status):-
                 Type = syntactic_concept_relevance,
                 K=1,
                 !, 
                 get_query_initial_conceptset(Q, C),
	   	     selected_set(Type, icr([],C,[]), [], KBID, L),                
                 icr_set(L, ICRk),
		     dig_consistent_check(elements(L), Status),
                 true.


pion_selection_function(Type, KBID, Q, K, _CurrentWorkingSet, _ICRk1, L, L, ICR1, Status):-
                (Type = syntactic_relevance; Type = syntactic),
                 K=1,
                 !,
      	     icr_set_element(Q, ICR),
      	     selected_set(Type, ICR, [], KBID, L),
 		     icr_set(L, ICR1),
		     dig_consistent_check(elements(L), Status),
                 true.



pion_selection_function(Type, KBID, _Query, K, CurrentWorkingSet, ICRk1, NewSelectedSet, NewWorkingSet, ICRk, Status):-
                (Type = syntactic_relevance; Type = syntactic; Type=syntactic_concept_relevance),
                 K>1,
                 !,
		    selected_set(Type, ICRk1, CurrentWorkingSet, KBID, NewSelectedSet),               
                ICRk1 =icr(IQ, CQ, RQ),
                icr_set(NewSelectedSet, icr(IP, CP, RP)),
                union(CurrentWorkingSet, NewSelectedSet, NewWorkingSet),
                union(IQ, IP, I),
		    union(CQ, CP, C),
		    union(RQ, RP, R),
                ICRk = icr(I,C,R),
                dig_consistent_check(elements(NewWorkingSet), Status),                
                true.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.10.0 added, 1.15.0 revision
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


pion_selection_function(Type, KBID, Q, K, _, _, L, L, ICRk, Status):-
                 Type = semantic_relevance_ngd,
                 K=1,
                 !, 
                 get_corekeyword(Q, CoreKeyword),
                 get_query_initial_conceptset(Q, C),
	   	     selected_set(semantic_relevance(ngd, CoreKeyword), icr([],C,[]), [], KBID, L),                         
                 icr_set(L, ICRk),
%                 format(user_error, 'NewWorkingSetL=~q, ICRk=~q~n', [L,ICRk]),       
		     dig_consistent_check(elements(L), Status),
%                 format(user_error,'L=~q~nK1 status=~w~n', [L,Status]),
                 true.



pion_selection_function(Type, KBID, Q, K, CurrentWorkingSet, ICRk1, NewSelectedSet, NewWorkingSet, ICRk, Status):-
                Type = semantic_relevance_ngd,
                K>1,
                !,
                get_corekeyword(Q, CoreKeyword), 
 % it should be improved here. getting corekeyword has been done  when k=1.

		    selected_set(semantic_relevance(ngd, CoreKeyword), ICRk1, CurrentWorkingSet, KBID, NewSelectedSet),               
 %               format(user_error, 'NewWorkingSet=~q, ICRk1=~q~n', [NewSelectedSet,ICRk1]),                  
                ICRk1 =icr(IQ, CQ, RQ),
                icr_set(NewSelectedSet, icr(IP, CP, RP)),
                union(CurrentWorkingSet, NewSelectedSet, NewWorkingSet),
                union(IQ, IP, I),
		    union(CQ, CP, C),
		    union(RQ, RP, R),
                ICRk = icr(I,C,R),
                dig_consistent_check(elements(NewWorkingSet), Status),
%                format(user_error, 'Status=~w, ICRk=~q~n', [Status, ICRk]),                
                true.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.4.0 added
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pion_selection_function(Type, KBID, Q, K, _, _, L, L, ICRk, Status):-
                 (Type = google_distance_min; Type = google_distance_max; Type= google_distance_average),
                 K=1,
                 !, 
                 get_query_initial_conceptset(Q, C),
	   	     selected_set(Type, icr([],C,[]), [], KBID, L),                         
                 icr_set(L, ICRk),
%                 format(user_error, 'NewWorkingSetL=~q, ICRk=~q~n', [L,ICRk]),       
		     dig_consistent_check(elements(L), Status),
%                 format(user_error,'L=~q~nK1 status=~w~n', [L,Status]),
                 true.

pion_selection_function(Type, KBID, _Query, K, CurrentWorkingSet, ICRk1, NewSelectedSet, NewWorkingSet, ICRk, Status):-
                (Type = google_distance_min; Type = google_distance_max; Type=google_distance_average),
                 K>1,
                 !,
%               format(user_error, 'K=~w, CurrentWorkingSet=~q~n', [K, CurrentWorkingSet]),
		    selected_set(Type, ICRk1, CurrentWorkingSet, KBID, NewSelectedSet),               
 %               format(user_error, 'NewWorkingSet=~q, ICRk1=~q~n', [NewSelectedSet,ICRk1]),                  
                ICRk1 =icr(IQ, CQ, RQ),
                icr_set(NewSelectedSet, icr(IP, CP, RP)),
                union(CurrentWorkingSet, NewSelectedSet, NewWorkingSet),
                union(IQ, IP, I),
		    union(CQ, CP, C),
		    union(RQ, RP, R),
                ICRk = icr(I,C,R),
                dig_consistent_check(elements(NewWorkingSet), Status),
%                format(user_error, 'Status=~w, ICRk=~q~n', [Status, ICRk]),                
                true.


pion_selection_function(Type, _KBID, _Query, _K, _CurrentWorkingSet, _ICRk1, NewSelectedSet, NewWorkingSet, ICRk, Status):-
            !,
		format(user_error, 'unsupported selection function ~w or parameter error~n', [Type]),
            Status =error,
            NewSelectedSet = [],
            NewWorkingSet = [],
            ICRk = icr([],[],[]).


get_query_initial_conceptset(Q, C):-
    Q = element(subsumes, _, L),
    dig_get_element(L, _E1, L1),
    dig_get_element(L1, E2, _),
    !,
    icr_set_element(E2, ICR),
    ICR=icr(_, C, _),
    true.

                         


%pion_relevance(+Type, +Element/+ICR,  +Element/+Elements/+ICR)

pion_relevance(syntactic_concept_relevance, X,Y):-!,
       direct_concept_relevant(X,Y).


pion_relevance(syntactic_relevance, X, Y):-!,
       relevant(X,Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.2.0 revised, v1.2.3 revised
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%pion_selection_function(+Type, +Query/PreSelectedSet, +CurrentSelectedSet, +KBID, -NewSelectedSet)

% syntactic relevance selection function
pion_selection_function(Type1, Q, P, KBID, L):-
      (Type1=syntactic_relevance; Type1=syntactic),
      !,
      Type = syntactic_relevance,
      icr_set_element(Q, icr(IQ, CQ, RQ)),
      icr_set(P, icr(IP, CP, RP)),
      union(IQ, IP, I),
	union(CQ, CP, C),
	union(RQ, RP, R),
	selected_set(Type, icr(I,C,R), P, KBID, L).


% deal with the initial selected set for the syntactic concept relevance
% selection function
%v1.2.3
pion_selection_function(Type, Q, P, KBID, L):-
      Type=syntactic_concept_relevance,
      P = [],
      !,
      icr_set_element(Q, icr(IQ,_,_)),
      selected_set(syntactic_relevance, icr(IQ,[],[]), [], KBID, L1),
%%      format(user_error, 'initial selected:~w~n',[L1]),
      icr_set(L1, ICR1),
	selected_set(Type, ICR1, L1, KBID, L2),
      append(L1, L2, L).
      

%v1.2.3
pion_selection_function(Type, Q, P, KBID, L):-
      Type=syntactic_concept_relevance,
      !,
      icr_set_element(Q, icr(IQ, CQ, RQ)),
      icr_set(P, icr(IP, CP, RP)),
      union(IQ, IP, I),
	union(CQ, CP, C),
	union(RQ, RP, R),
	selected_set(Type, icr(I,C,R), P, KBID, L).


pion_selection_function(Type, _Q, _P, _KBID, _L):-
      format(user_error, 'undefined selection function type: ~w~n',[Type]).


%selected_set(+ICRSetofCurrentWorkingSet, +CurrentWorkingSet, +KBID, -SelectedSet)
selected_set(syntactic_relevance, ICR, P, KBID, L):-
	bagof(E, (dig_db_element(KBID, E),\+ member(E, P), relevant(E, ICR)), L1),
      !,
      list_to_set(L1, L),
      true.


%  version 1.2.3 added
selected_set(syntactic_concept_relevance, ICR, P, KBID, L):-
     bagof(E, (dig_db_element(KBID, E),\+ member(E, P), direct_concept_relevant(E, ICR)), L1),
     !,
     list_to_set(L1, L),
     true.

%  version 1.15.0 added
selected_set(semantic_relevance(ngd, CoreKeyword), ICR, P, KBID, L3):-
     bagof(E, (dig_db_element(KBID, E),\+ member(E, P), direct_concept_relevant(E, ICR)), L1), % apply direct concept relevance first
     !,
     list_to_set(L1, L),
     ICR=icr(I, C, R),
     append(I, C,ICL),
     append(ICL, R, ICRL),
     google_distance_relevant(semantic_relevance(ngd, CoreKeyword), ICRL, L, L3, _Distance),
     true.




%  version 1.4 added, 1.6, 1.7, 1.15 modified
% for google distance without applying syntactic relevance selection first
selected_set(Type, ICR, P, KBID, L3):-
        bagof(E, (dig_db_element(KBID, E),\+ member(E, P)), L1),
        list_to_set(L1,L),
         ICR=icr(I, C, R),
         append(I, C,ICL),
         append(ICL, R, ICRL),
         google_distance_relevant(Type, ICRL, L, L3, _Distance),
%        format(user_error, 'find L3=~q, Distance=~q~n', [L3, Distance]),
%        format(user_error, 'find Distance=~q~n', [Distance]),
        true.


% selected set as emptyset if the above fails.
% version 1.2.3 added
selected_set(_Type, _L, _P, _KBID, L):-
   L= [].

% version 1.10 added, 1.15 revision
google_distance_relevant_corekeyword(_CoreKeyword, [], [], 1):-!.


google_distance_relevant_corekeyword(CoreKeyword, [E1|L1],L3, D):-
     get_axiom_distance(CoreKeyword, E1,  google, D1),
     !,
     google_distance_relevant_corekeyword(CoreKeyword, L1, L2, D2),
     (D1 < D2 -> (L3=[E1], D=D1); (D1=D2-> (L3=[E1|L2], D=D1); (L3=L2, D=D2))),
     true.


% version 1.10 added, 1.15 revision
google_distance_relevant(Type, _ICRL, L, L3, D):-
   Type = semantic_relevance(ngd, CoreKeyword), 
   !,
   google_distance_relevant_corekeyword(CoreKeyword, L, L3, D).



%  version 1.4 added, 1.6, 1.7 revised


google_distance_relevant(_Type,  _ICRL, [], [], 1).


google_distance_relevant(Type,  ICRL,  [E1|L], L3, D):-
     google_distance(Type, ICRL, E1, D1),
     !,
%     format(user_error, 'google_distance find: E1=~w, D1=~w***~n', [E1,D1]),
     google_distance_relevant(Type,  ICRL,  L, L2,D2),
     (D1 < D2 -> (L3=[E1], D=D1); (D1=D2-> (L3=[E1|L2], D=D1); (L3=L2, D=D2))),
      true.
      
google_distance(Type, ICRL, E, Distance):-
      icr_set_element(E, icr(IE, CE, RE)),
      append(IE, CE, L1),
      append(L1,RE, L), 
%      format(user_error, '***E=~q~nL=~w~n', [E,L]),
      google_distance_list(Type, ICRL, L, Distance).

google_distance_list(Type, ICRL, [C], D):-
    google_distance_element(Type, ICRL, C, D).

google_distance_list(Type, ICRL, [C|L], D):- 
    google_distance_element(Type, ICRL, C, D1),
    google_distance_list(Type, ICRL, L, D2),
    (D1< D2 -> D = D1; D=D2),
    true.

google_distance_element(google_distance_min, [], _C, 1):-
    !.

google_distance_element(google_distance_min, [C|_L], C, 0):-
    !.

google_distance_element(google_distance_min, [C1|_L], C2, 0):-
    dig_get_shorted_name(C1, C11),
    dig_get_shorted_name(C2, C22),
    C11 = C22,
    !.


google_distance_element(google_distance_min, [C01|L], C0, D):-
%         format(user_error, 'C01=~w,C0=~w~n', [C01, C0]),
         dig_get_shorted_name(C01, C1),
         dig_get_shorted_name(C0, C),
%       format(user_error, 'ngd* C1=~q, C=~w, ', [C1,C]),
        (ngd(C1, C, D0)-> atom_number(D0, D1); D1 is 1),     
%       format(user_error, 'D0=~q~n', [D0]),            
        google_distance_element(google_distance_min, L, C, D2),
        (D1 < D2 -> D= D1; D=D2),
       true.
      
    

%  version 1.5.3 added

google_distance_element(google_distance_max, [], _C, 0):-
    !.

google_distance_element(google_distance_max, [C|_L], C, 0):-
    !.

google_distance_element(google_distance_max, [C1|_L], C2, 0):-
    dig_get_shorted_name(C1, C11),
    dig_get_shorted_name(C2, C22),
    C11 = C22,
    !.


google_distance_element(google_distance_max, [C01|L], C0, D):-
         dig_get_shorted_name(C01, C1),
         dig_get_shorted_name(C0, C),
%       format(user_error, 'ngd* C1=~q, C=~q,', [C1,C]),
        (ngd(C1, C, D0)-> atom_number(D0, D1); D1 is 1),     
%         format(user_error, 'D1=~q,', [D1]),           
        google_distance_element(google_distance_max, L, C, D2),
%        format(user_error, 'D2=~q~n', [D2]),
        (D2 < D1 -> D= D1; D=D2),
       true.

google_distance_element(google_distance_sum, [], _C, 0):-
    !.

google_distance_element(google_distance_sum, [C|_L], C, 0):-
    !.

google_distance_element(google_distance_sum, [C01|L], C0, D):-
         dig_get_shorted_name(C01, C1),
         dig_get_shorted_name(C0, C),
        (ngd(C1, C, D0)-> atom_number(D0, D1); D1 is 1),                  
        google_distance_element(google_distance_sum, L, C, D2),
        D is D1 + D2,
        true.

      
    
google_distance_element(google_distance_average, L, C, D):-
    length(L, N),
    google_distance_element(google_distance_sum, L, C, Sum),
    D is Sum/N.    



%relevant(ICR1, ICR2): two ICR are relevant
relevant(icr(IE,CE, RE), icr(I,C, R)):-
	(intersection(I, IE, [_I|_]); intersection(C, CE, [_C|_]); intersection(R, RE, [_R|_])),!.

%relevant(Element, ICR): the element is relevant to the ICR.
relevant(E, icr(I,C,R)):-
	icr_set_element(E, icr(IE, CE, RE)),
	(intersection(I, IE, [_I|_]); intersection(C, CE, [_C|_]); intersection(R, RE, [_R|_])).

%get the -ICR for an element +Set 
icr_set([], icr([],[],[])):-!.
icr_set([H|T], icr(I2,C2, R2)):-
			icr_set_element(H, icr(I,C, R)),
			icr_set(T, icr(I1,C1, R1)),
			union(I,I1,I2),
			union(C,C1,C2),
			union(R,R1,R2).



%get the -ICR set for an +Element 
icr_set_element(E, ICR):-
	(E= element(individual, [name=I],_); E=element(defindividual,[name=I],_)),
	!,
	ICR=icr([I], [], []).

icr_set_element(E, ICR):-
	(E= element(catom, [name=C],_); E=element(defconcept, [name=C],_)),
      \+ stop_set(c, C),
	!,
	ICR=icr([], [C], []).

icr_set_element(E, ICR):-
	(E= element(ratom, [name=R],_); E=element(defrole, [name=R],_)),
	!,
	ICR=icr([], [], [R]).

icr_set_element(E, ICR):-
	E= element(_, _X, L),
	!,
	icr_set(L, ICR).


icr_set_element(_,icr([],[],[])):-!.

% stop set defines a set of names which should be ignored on ICR sets.

stop_set(c, 'http://www.w3.org/2002/07/owl#Thing').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.2.3 added
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%direct_concept_relevant(+E1, +E2)

/* version 1.2.3
direct_concept_relevant(E1, E2):-
   E1 = element(impliesc, _X, L1),
   E2 = element(_,_,_),
   !,
   E3 = element(catom, [name=A], _), 
   dig_get_element(L1, E3, _),
   icr_set_element(E2, icr(_I, C, _R)),
   member(A, C).

direct_concept_relevant(E1, E2):-
   E2 = icr(_,C,_),
   E1 =element(impliesc, _X, L1),
   !,
   E3 = element(catom, [name=A], _), 
   dig_get_element(L1, E3, _),
   member(A, C).

direct_concept_relevant(E1, E2):-
   (E1 =element(equalc, _, L); E1 = element(disjoint, _, L)),
   E2 = element(_,_,_),
   member(element(catom, [name=A], _), L),
   icr_set_element(E2, icr(_I, C, _R)),
   member(A, C).

direct_concept_relevant(E1, E2):-
   (E1 =element(equalc, _, L); E1 = element(disjoint, _, L)),
   E2 = icr(_,C,_),
   member(element(catom, [name=A], _), L),
   member(A, C).


*/

% version 1.9 revision. 
% the new definition of  concept relevance which allows
% foldable axioms

direct_concept_relevant(E1, E2):-
   E1 =element(impliesc, _X, L1),
   E2 = icr(_,C2,_),
   !,
   dig_get_element(L1, E3, _),
   icr_set_element(E3, icr(_, C1, _)),
   member(A, C1),
   member(A, C2).

direct_concept_relevant(E1, E2):-
   E1 = element(impliesc, _X, L1),
   E2 = element(_,_,_),
   !,
   dig_get_element(L1, E3, _),
   icr_set_element(E3, icr(_, C1, _)),
   icr_set_element(E2, icr(_, C2, _)),
   member(A, C1),
   member(A, C2).



direct_concept_relevant(E1, E2):-
   (E1 =element(equalc, _, L); E1 = element(disjoint, _, L)),
   E2 = element(_,_,_),
   !,
   icr_set(L, icr(_,C1,_)),
   icr_set_element(E2, icr(_I, C2, _R)),
   member(A, C1),
   member(A, C2).

direct_concept_relevant(E1, E2):-
   (E1 =element(equalc, _, L); E1 = element(disjoint, _, L)),
   E2 = icr(_,C2,_),
   !,
   icr_set(L, icr(_,C1,_)),
   member(A, C1),
   member(A, C2).


%  end of version 1.9 revision







direct_concept_relevant(E1, L2):-
   is_list(L2),
   !,
   icr_set(L2, ICR),
   direct_concept_relevant(E1, ICR).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  PION naming convertion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pion_icr_set_element(X,Y):-
		icr_set_element(X,Y).

pion_icr_set(X,Y):- icr_set(X,Y).
         	
pion_direct_concept_relevant(X,Y):-
	direct_concept_relevant(X,Y).

pion_relevant(X,Y):-
	relevant(X,Y).		


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% end of pion_selection.pl 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

